/* ========================================
 *
 * Software example by Ralf S. Mayer, 2018
 * reviewed and translated by Holger Wech
 * All Rights Reserved
 *
 * ========================================
*/
#include <project.h>
#include <stdio.h>
#include <string.h>     // eg. strtok
#include <stdlib.h>     // eg. atoi, atol, atof

#include "WS2812.h"

/** Define polarity depending on used type of RGB-LED */
#define RGBLED_ON   ( 1u )  //!< RGB-LED common cathode on
#define RGBLED_OFF  ( 0u )  //!< RGB-LED common cathode off
//#define RGBLED_ON   ( 0u )  //!< RGB-LED common anode on
//#define RGBLED_OFF  ( 1u )  //!< RGB-LED common anode off

#define LED_R(x) ( Pin_LED_R_Write(x) ) //!< RGB-LED Pin R
#define LED_G(x) ( Pin_LED_G_Write(x) ) //!< RGB-LED Pin G
#define LED_B(x) ( Pin_LED_B_Write(x) ) //!< RGB-LED Pin B

/**
 * \def USERLED_ONOFF
 * Switch user LED on or off.
 *
 * Pin to be used depending on board.
 * @param x  1: on, 0: off
 */
#define USERLED_ONOFF(x) ( Pin_USR_LED_Write(x) )
#define USRLED_ON   ( 1u )  //!< user-LED common cathode on
#define USRLED_OFF  ( 0u )  //!< user-LED common cathode off

// +++ globals +++
extern volatile  uint32_t u32DelayDemoMode;    // run demo striplight show if 0 

int16 aRGB[4];          // ADC values
int16 rRGB[3];          // normalized ADC values
int16 sRGB[3];          // saturated RGB values
float rgb[3];           // float values
float h, s, v;          // H, S, V values
uint8 flagContd = 0;    //!< flag for continous display
uint8 flagIntens = 0;   //!< flag for intense display
int16 brightn;          // brightness value
char cInput = 0;        // value for keyboard input

char buffer[128];       //!< buffer (global) for sprintf
uint8 flagVerb = 0;     //!< verbosity flag

uint8 flagUART_RX = 0;  //!< flag (global) itr has occured
static char cRx = 0;    //!< character from UART, visible within main.c

#define NofSamples 10         //!< number of samples for color measurement
typedef int16(*Fadc_t)(void); //!< function type ADC_function, e.g. _GetResult16
//#define MAX_COLOR_VALUE 100   //!< maximum normalized color value
#define MAX_COLOR_VALUE 255   //!< maximum normalized color value
#define MEAN_COLOR_VALUE (MAX_COLOR_VALUE/3) //!< mean color value

/** normalized ADC values for a white object and initialization */
static int16 whiteRGB[3] = {
    MEAN_COLOR_VALUE, MEAN_COLOR_VALUE, MEAN_COLOR_VALUE
};   

//#define ADC_PD_MAX 5500 //!< assumed maximum value of ADC for color measurement
#define ADC_PD_MAX (5500/100) //!< normalization factor for ADC (0 ... 100)

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++ the display RGB LED does not have                  +++
// +++ the same (physiological) brightness for each color +++ 
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#define R_CAL   116     //!< display RGB(R) LED start value red
#define G_CAL    47     //!< display RGB(G) LED start value green
#define B_CAL    47     //!< display RGB(B) LED start value blue

/** Display RGB calibration values, initialization */
static uint8 dispCalRGB[3] = { R_CAL, G_CAL, B_CAL };

/** Display test values */
static int16 wTestRGB[3] = { 
    MEAN_COLOR_VALUE, MEAN_COLOR_VALUE, MEAN_COLOR_VALUE 
};

#define BRIGHTN_THRESHOLD 150    //!< brightness threshold for continous display
#define SATURATION_THRESHOLD 0.2 //!< saturation threshold for color management

/** color names in HSV 60 degree steps (except gray)*/
static char* colName[] = 
{ "red", "yellow", "green", "cyan", "blue", "magenta", "gray" };

// ++++++++++++++++++++++++++++
// +++ forward declarations +++
// ++++++++++++++++++++++++++++

/** @brief Conversion RGB to HSV format */
void RGBtoHSV( float r, float g, float b, float *ph, float *ps, float *pv);

/** @brief Normalize rgb float Value from RGB */
float* normRGB( int16* vRGB, float* rgb );

/** @brief Get color name from the Hue wheel in the HSV system */
char* getColorName( float h, float s );

/** @brief Saturate RGB values for more intense hue in display */
int16* satRGB( int16* vRGB, int16* sRGB );

/** 
 * UART RX interrupt callback.
 *
 * Collect a characters from UART into a global input variable.
 * @see cRx
 */
CY_ISR(isr_UART_RX)
{
    cRx = UART_GetChar();     // copy char    
}

/**
 * This function will control the RGB illumination and will take samples of the photo-diode
 * Finally it calculates the average value calculated out of NofSamples measurements
 * @see NofSamples
 * @param vRGB Array[4] including {R, G, B, ambient light}
 * @param pfADC Pointer to function ADC_SAR_PD_GetResult16()
 * @return Pointer to vRGB, 0 if error
 *
 * More information on the Transimpedance amplifier (TIA) built in PSoC5LP
 * @see https://en.wikipedia.org/wiki/Transimpedance_amplifier
 * @see TIA_PD
 */
int16* measureColor( int16* vRGB, Fadc_t pfADC ) {
    if ( vRGB == 0 ) {          // invalid pointer        
        return 0;
    }
    
    int16 vCol[NofSamples][4];
    int32 vRGBs[4];             // Array Summenvariable
    uint32 wait = 5;            // wait 5ms
    
    // Switch off external illumination 
    
    LED_R( RGBLED_OFF );
    LED_G( RGBLED_OFF );
    LED_B( RGBLED_OFF );
    
    // RGB Measurement
    
    uint8 i;
    for ( i = 0; i < NofSamples; i++ ) {
        CyDelay( wait );
        vCol[i][3] = pfADC();   // Measure ambient light (All RGB LEDs off)
        LED_R( RGBLED_ON );
        CyDelay( wait );
        vCol[i][0] = pfADC();   // Measure red light reflection  
        LED_R( RGBLED_OFF );
        LED_G( RGBLED_ON );
        CyDelay( wait );
        vCol[i][1] = pfADC();   // Measure green light reflection
        LED_G( RGBLED_OFF );
        LED_B( RGBLED_ON );
        CyDelay( wait );
        vCol[i][2] = pfADC();   // Measure blue light reflection
        LED_B( RGBLED_OFF );
    }
    
    // Average calculation
    vRGBs[0] = vRGBs[1] = vRGBs[2] = vRGBs[3] = 0;    // initial value
    for ( i = 0; i < NofSamples; i++ ) {
        // sum up values
        vRGBs[0] += vCol[i][0] -  vCol[i][3];     // R value - background
        vRGBs[1] += vCol[i][1] -  vCol[i][3];     // G value - background
        vRGBs[2] += vCol[i][2] -  vCol[i][3];     // B value - background
        vRGBs[3] += vCol[i][3];                   // background
    }
    
    // Average
    vRGB[0] = vRGBs[0] / NofSamples;
    vRGB[1] = vRGBs[1] / NofSamples;
    vRGB[2] = vRGBs[2] / NofSamples;
    vRGB[3] = vRGBs[3] / NofSamples;
    
    return vRGB;
} // end measureColor

/**
 * Calculation of relative RGB-values
 * 
 * @param cRGB Array[3] with {R, G, B} relative
 * @param vRGB Array[4] with averaged {R, G, B, ambient}
 * @see measureColor
 * @return absolute brightness = Sum( R, G, B ) from input or 0 if error
 *         Return of normalized values in cRGB
 */
int16 calcColor( int16* cRGB, int16* vRGB ) {
    if ( ( vRGB == 0 ) || ( cRGB == 0 ) ) {     // invalid pointer        
        return 0;
    }

    int16  rgbAbs = vRGB[0] + vRGB[1] + vRGB[2];       // absolute luminosity

    int i;
    // calculate relative RGB value * MAX_COLOR_VALUE
    for ( i = 0; i < 3; i++ ) {
        cRGB[i] = ( vRGB[i] * MAX_COLOR_VALUE ) / rgbAbs;
    }

    return rgbAbs;
}

/**
 * Calculation of RGB-values white calibrated.
 * using global calibration value whiteRGB
 * @see whiteRGB
 * @see calcColor()
 * 
 * @param cRGB Array[3] with {R, G, B} relative, in/out.
 * @param vRGB Array[4] with average {R, G, B, ambient}
 * @see measureColor
 * @return absolute brightness = Sum( R, G, B ) from input or 0 if error
 *         Return of normalized values in cRGB
 */
int16 calcColorCal( int16* cRGB, int16* vRGB ) {
    if ( ( vRGB == 0 ) || ( cRGB == 0 ) ) {     // invalid pointer        
        return 0;
    }
    
    int16 calRGB[3];   // white calibrated values

    int i;
    // calibrate to white balance
    for ( i = 0; i < 3; i++ ) {
        calRGB[i] = vRGB[i] * MEAN_COLOR_VALUE / whiteRGB[i];
    }
    
    return calcColor( cRGB, calRGB );
}

/**
 * Display sensored RGB values with the RGB-LED
 * 
 * @param cRGB Array[3] mit {R, G, B} relativ (0 ... 100)
 */
void displayColor( int16* cRGB ) {
    // Output sampled RGB value on StripLight/WS2812
    
    uint32_t u32Color;
    u32Color = (((cRGB[0] * dispCalRGB[0] / 100 ) << 8) & 0xFF00) + ((cRGB[1] * dispCalRGB[1] / 100 ) & 0xFF) + (((cRGB[2] * dispCalRGB[2] / 100)<<16)&0x00FF0000); 
    StripLights_MemClear(u32Color); // Set all LEDs to calculated value
 
    //    For debugging only: Display result on terminal
    //    
    //    if (flagVerb)
    //    {
    //        sprintf( buffer, "cRGB:       %02x %02x %02x\r\n", cRGB[0], cRGB[1], cRGB[2]);
    //        UART_PutString(buffer);
    //        sprintf( buffer, "dispCalRGB: %02x %02x %02x\r\n", dispCalRGB[0], dispCalRGB[1], dispCalRGB[2]);
    //        UART_PutString(buffer);
    //        sprintf( buffer, "Display-Color [--bbrrgg]: 0x%08x\r\n", u32Color); // WS2812 format
    //        UART_PutString(buffer);
    //    }
}

/**
 * Complete Color Sampling process (sample, calculate, output)
 * 
 * @param none
 */
void RGBcolorSampling(void)
{
    measureColor( aRGB, ADC_SAR_PD_GetResult16 );
    brightn = calcColorCal( rRGB, aRGB );
    // transform RGB to HSV color system
    normRGB( rRGB, rgb );
    RGBtoHSV( rgb[0], rgb[1], rgb[2], &h, &s, &v );
    sprintf( buffer, "Color is < %s >\r\n",
        getColorName( h, s ) );
    UART_PutString( buffer );
    // display color with RGB LED 
    if ( flagIntens ) {     // intense color display
        if ( s >= SATURATION_THRESHOLD ) {// saturation threshold
            displayColor( satRGB( rRGB, sRGB ) );
        }
        else                // normal color display
            displayColor( rRGB );
    }
    else {                  // normal color display
        displayColor( rRGB );
    }
    // additional output
    if ( flagVerb ) 
    {   // verbose on
        sprintf( buffer, "WBcal  r: %3d, g: %3d, b: %3d, dark: %3d",
            aRGB[0], aRGB[1], aRGB[2], aRGB[3] );
        UART_PutString( buffer );
        sprintf( buffer, " -> brightn: %3d: R: %3d, G: %3d, B: %3d -> (HSV %3d/100 %2d/100 %2d/100)\r\n",
            brightn, rRGB[0], rRGB[1], rRGB[2],
            (int) h, (int) (s*100.), (int) (v*100.) );
        UART_PutString( buffer );
    } 
}

/**
 * Complete White Sampling process (sample, calculate, output)
 * requires a white reference object placed below the sensor
 * 
 * @param none
 */
void RGBwhiteSampling()
{
    //UART_PutString( "Present white object below sensor ...\r\n" );
    measureColor( aRGB, ADC_SAR_PD_GetResult16 );
    brightn = calcColor( whiteRGB, aRGB );
    sprintf( buffer, "White balance: -> brightness: %3d: R: %3d, G: %3d, B: %3d\r\n",
        brightn, 
        whiteRGB[0], whiteRGB[1], whiteRGB[2] );
    UART_PutString( buffer );
}

/**
 * Display 'help' information, mainly all available key commands
 * 
 * @param none
 */
void prtHelp()
{
    UART_PutString("\r\nHelp menu:\r\n\th,H help\r\n");
    UART_PutString("\tw: white balance, present a white object to calibrate\r\n");
    UART_PutString("\ts: Sample object below photo diode (white balanced)\r\n");
    UART_PutString("\tS: Sample object below photo diode (not white balanced)\r\n");
    UART_PutString("\tC: Continuous sampling on\r\n");
    UART_PutString("\tc: Continuous sampling off\r\n");
    UART_PutString("\tX: Switch RGB LED all on (white)\r\n");
    UART_PutString("\tx: Switch RGB LED all off\r\n");
    UART_PutString("\t1: Switch RGB LED to red\r\n");
    UART_PutString("\t2: Switch RGB LED to green\r\n");
    UART_PutString("\t3: Switch RGB LED to blue\r\n");
    UART_PutString("\t0: Switch RGB LED all off\r\n");
    UART_PutString("\tV: Verbose output on\r\n");
    UART_PutString("\tv: Verbose output off\r\n");
    UART_PutString("\tt: Test WS2812 LED string\r\n");
    UART_PutString("\r\n");
}

/**
 * RGB-Color-Sensor based on PSoC5LP microcontroller
 * using CY8CKIT-059 or Sparkfun FreeSoC2
 *
 * Used peripherals: UART, TIA, ADC, GPIO 
 * Optional: SysTick, Timer, Striplights WS2812
 *
 * Control by UART, for help and function
 * @see prtHelp()
 *
 *
 * @author Ralf S. Mayer (Co-author Holger Wech)
 * @date 2018-03-29
 * @version 1.3
 *
 *
 * Additonal background information (German language):
 * @see http://www.fseitz.de/blog/index.php?/archives/112-Helligkeit-von-Farben-des-RGB-Farbraums-berechnen.html
 * @see http://www.kurztutorial.info/programme/helligkeit-farbe-berechnen/rgb-bright-rechner.htm
 * @see https://wisotop.de/hsv-und-hsl-farbmodell.php
 */
int main()
{
    CyGlobalIntEnable; /* Enable global interrupts. */

    // Optional: StripLights initialization
    
    initWS2812();
    
    // UART initialization
    
    irq_UART_RX_StartEx(isr_UART_RX);   // enable itr uart
    UART_Start();                       // start uart

    UART_PutString("\r\nWelcome to Cypress PSoC5LP RGB Sensor");
        
    prtHelp();                          // help text
  
    // Photo Diode ADC Initialization
    TIA_PD_Start();             // Start TIA for photo diode
    ADC_SAR_PD_Start();         // Start ADC
    ADC_SAR_PD_StartConvert();  // Start ADC conversion

    // set intense color on
    flagIntens = 1;

    u32DelayDemoMode = 10;  // do not enter demo mode for x*100ms
    showWS2812(2);          // Activate StripLights (1: Rotate LEDs) - just for fun
    
    // Automatic calibration after reset requires to have a white background always (at least after reset) below the sensor
    UART_PutString( "Present white object below sensor ...\r\n" );
    RGBwhiteSampling();     
    UART_PutString("Automatic white calibration done!\r\n"); 
    
    UART_PutString("System ready!\r\n\r\n");
    
    // Main loop
    for(;;)     // forever
    {
        /* Place your application code here. */
        // uart char received?
        
        if (SampleSW_Read() == 0)
        {
            u32DelayDemoMode = 100; // do not enter demo mode for x*100ms
            showWS2812(2);          // Activate StripLights (1: Rotate LEDs) - just for fun
            RGBcolorSampling();     // activate sensor and display result    
        }
        
        if ( cRx != 0 ) {           // char received
            irq_UART_RX_Disable();  // disable Itr
            cInput = cRx;           // copy char
            cRx = 0;                // reset
            irq_UART_RX_Enable();   // enable Itr
        }

        if ( cInput != 0 ) {        // has input
            UART_PutChar(cInput);   // print input key
            switch ( cInput ) {     // handle input
                
                // Help menue
                
                case 'h':
                case 'H':
                    UART_PutString(" ->\r\n");
                    prtHelp();      // help text
                    UART_PutString( "\r\nMake your selection:\r\n" );
                    break;
                    
                // Test purpose: control of RGB LED    
                case 'X':
                    UART_PutString(" -> Switch RGB LED to white (all LEDs on)\r\n" );
                    LED_R( RGBLED_ON );     // on
                    LED_G( RGBLED_ON );     // on
                    LED_B( RGBLED_ON );     // on
                    break;
                case '1':
                    UART_PutString(" -> Switch RGB LED to red\r\n" );
                    LED_R( RGBLED_ON );      // on
                    LED_G( RGBLED_OFF );     // on
                    LED_B( RGBLED_OFF );     // on
                    break;
                case '2':
                    UART_PutString(" -> Switch RGB LED to green\r\n" );
                    LED_R( RGBLED_OFF );     // on
                    LED_G( RGBLED_ON );      // on
                    LED_B( RGBLED_OFF );     // on
                    break;
                case '3':
                    
                    UART_PutString(" -> Switch RGB LED to blue\r\n" );
                    LED_R( RGBLED_OFF );     // on
                    LED_G( RGBLED_OFF );     // on
                    LED_B( RGBLED_ON );      // on
                    break;
                case '0':
                case 'x':
                    UART_PutString(" -> Switch RGB LED off\r\n" );
                    LED_R( RGBLED_OFF );     // off
                    LED_G( RGBLED_OFF );     // off
                    LED_B( RGBLED_OFF );     // off
                    break;
                    
                // Start/Stop repetitive color sampling 
                    
                case 'C':       // start continous display (flag handled below)
                    UART_PutString(" -> Repetitive mode is on\r\n" );
                    flagContd = 1;
                    break;
                
                case 'c':       // stop continous display (flag handled below)
                    UART_PutString(" -> Repetitive mode is off\r\n" );
                    flagContd = 0;
                    break;
                  
                // Switch intensity
                    
                case 'I':       // intense display on
                    flagIntens = 1;
                    UART_PutString( " -> Intense color display ON\r\n" );
                    break;

                case 'i':       // intense display off
                    flagIntens = 0;
                    UART_PutString( " -> Intense color display OFF\r\n" );
                    break;
                    
                // Color sampling
                    
                case 's':   // white calibrated values
                    UART_PutString(" -> Sample object (using white calibration)\r\n" );
                    u32DelayDemoMode = 200;
                    StripLights_MemClear(StripLights_BLACK);
                    RGBcolorSampling(); // activate sensor and display result 
                    break;                   
                
                case 'S':   // not white calibrated
                    UART_PutString(" -> Sample object (not using white calibration)\r\n" );
                    measureColor( aRGB, ADC_SAR_PD_GetResult16 );
                    sprintf( buffer, "uncal  r: %3d, g: %3d, b: %3d, dark: %3d",
                        aRGB[0], aRGB[1], aRGB[2], aRGB[3] );
                    UART_PutString( buffer );
                    brightn = calcColor( rRGB, aRGB );
                    sprintf( buffer, "\t -> brightn: %3d: R: %3d, G: %3d, B: %3d\r\n",
                        brightn, rRGB[0], rRGB[1], rRGB[2] );
                    UART_PutString( buffer );
                    displayColor( rRGB );   // Output result
                    break;                   
                
                case 'w':                  // white balance
                    UART_PutString(" -> White calibration (white reference must be below sensor!)\r\n" );
                    RGBwhiteSampling();    // activate sensor and take white reference value
                    break;
                    
                // Verbose mode
                    
                case 'V':       // verbose on
                    UART_PutString(" -> Verbose mode is on\r\n");
                    flagVerb = 1;
                    break;
                case 'v':       // verbose off
                    UART_PutString(" -> Verbose mode is off\r\n");
                    flagVerb = 0;
                    break;
                
                // Individual calibration
                    
                // +++ for display calibration +++
                case 'R':   // increment RED
                    UART_PutString(" -> Increment red\r\n");
                    dispCalRGB[0]++;
                    displayColor( wTestRGB );
                    break;
                case 'r':   // decrement RED
                    UART_PutString(" -> Decrement red\r\n");
                    dispCalRGB[0]--;
                    displayColor( wTestRGB );
                    break;
                case 'G':   // increment GREEN
                    UART_PutString(" -> Increment green\r\n");
                    dispCalRGB[1]++;
                    displayColor( wTestRGB );
                    break;
                case 'g':   // decrement GREEN
                    UART_PutString(" -> Decrement green\r\n");
                    dispCalRGB[1]--;
                    displayColor( wTestRGB );
                    break;
                case 'B':   // increment BLUE
                    UART_PutString(" -> Increment blue\r\n");
                    dispCalRGB[2]++;
                    displayColor( wTestRGB );
                    break;
                case 'b':   // decrement BLUE
                    UART_PutString(" -> Decrement blue\r\n");
                    dispCalRGB[2]--;
                    displayColor( wTestRGB );
                    break;
                    
                // Show values
                    
                case 'k':   // show display calibration
                    UART_PutString(" -> Calibration values: " );
                    displayColor( wTestRGB );
                    sprintf( buffer, "disp cal RGB: %d %d %d\r\n",
                    dispCalRGB[0], dispCalRGB[1], dispCalRGB[2] );
                    UART_PutString( buffer );
                    break;
                
                // Test WS2812 Striplights
                    
                case 't':   // LED test of NeoPixel(WS2812B)
                    u32DelayDemoMode = 100;
                    UART_PutString(" -> Observe the NeoPixel(WS2812B) ... " );
                    testWS2812();
                    UART_PutString("done!\r\n");
                    break;    
                  
                // Unknown command
                    
                default:    // repeat char
                    UART_PutChar( cInput );
                    UART_PutString(" -> Unknown command!\r\n" );
                    break;
            } // end switch
            cInput = 0;             // reset, don't forget!
        }
        
        // Handle continous diplay flag (calibrated)
        if ( flagContd ) {
            measureColor( aRGB, ADC_SAR_PD_GetResult16 );
            brightn = calcColorCal( rRGB, aRGB );
            
            if ( brightn > aRGB[3] ) {              // relative threshhold
                if ( flagIntens ) {     // intense color display
                    // transform RGB to HSV color system
                    normRGB( rRGB, rgb );
                    RGBtoHSV( rgb[0], rgb[1], rgb[2], &h, &s, &v );
                    sprintf( buffer, "Color is < %s >\r\n",
                    getColorName( h, s ) );
                    UART_PutString( buffer );
                    CyDelay(500);                       // Delay to allow readable UART output
                    if ( s >= SATURATION_THRESHOLD ) {// saturation threshold
                        displayColor( satRGB( rRGB, sRGB ) );
                    }
                    else                // normal color display
                        displayColor( rRGB );
                }
                else {                  // normal color display
                    displayColor( rRGB );
                }
            }
            else {
                int16 nullRGB[3] = { 0, 0, 0 };
                displayColor( nullRGB );
            }
        }
                
    } // end for ever
} // end main

// +++ Helper Functions +++

/** 
 * \def MIN
 * Return Minimum of two values
 * @param a
 * @param b
 * @return minimum( a, b )
 */
#define MIN(a,b) ( (a<b) ? a:b ) 
/** 
 * \def MAX
 * Return Maximum of two values
 * @param a
 * @param b
 * @return maximum( a, b )
 */
#define MAX(a,b) ( (a>b) ? a:b ) 

/** 
 * \def MIN3
 * Return Minimum of three values
 * @param a
 * @param b
 * @param c
 * @return minimum( a, b, c )
 */
#define MIN3(a,b,c) ( MIN( MIN(a,b), c ))   //!< Minimum aus drei Zahlen
/** 
 * \def MAX3
 * Return Maximum of three values
 * @param a
 * @param b
 * @param c
 * @return maximum( a, b, c )
 */
#define MAX3(a,b,c) ( MAX( MAX(a,b), c ))   //!< Maximum aus drei Zahlen

/**
 * Conversion RGB to HSV
 * @author Wisotop, modified Ralf S. Mayer
 * @see https://wisotop.de/rgb-nach-hsv.php
 *
 * @param r R value [0,1]
 * @param g G value [0,1]
 * @param b g value [0,1]
 * @param ph Pointer to Hue, [0,360]
 * @param ps Pointer to Saturation, [0,1]
 * @param pv Pointer to Value, [0,1]
 * 
 */
void RGBtoHSV( float r, float g, float b, float *ph, float *ps, float *pv) {
   float min, max, delta;
   min = MIN3( r, g, b );
   max = MAX3( r, g, b );
   *pv = max;                        // v
   delta = max - min;
   if( max != 0 ) *ps = delta / max;
   else {                           // r = g = b = 0
      *ps = 0; *ph = -1; return;
   }
   if (max == min) {                //Grey
      *ph = 0; *ps = 0; return;
   }
   if( r == max )
      *ph = ( g - b ) / delta;       // between Yellow and Magenta
   else if( g == max )
      *ph = 2 + ( b - r ) / delta;   // between Cyan and Gelb
   else
      *ph = 4 + ( r - g ) / delta;   // between Magenta and Cyan
      *ph *= 60;                     // degrees
   if( *ph < 0 )
      *ph += 360;
}

/**
 * Normalize rgb float value from RGB
 * @param vRGB Integer RGB-value [0, MAX_COLOR_VALUE]
 * @param rgb normierte rgb-value [0,1]
 * @return Pointer to rgb-Array
 */
float* normRGB( int16* vRGB, float* rgb ) {
    int i;
    for ( i = 0; i < 3; i++ ) {
        rgb[i] = (float) vRGB[i] / MAX_COLOR_VALUE;
    }
    
    return rgb;
}

/**
 * Saturate RGB values for more intense hue in display.
 * Discriminate by the minimum of all values.
 *
 * Cave! Does not make sens for gray, white or less saturated colors!
 * @param vRGB integer values RGB [0, MAX_COLOR_VALUE]
 * @param sRGB saturated RGB values [0, MAX_COLOR_VALUE]
 * @return pointer to sRGB
 */
int16* satRGB( int16* vRGB, int16* sRGB ) {
    // get minimum value from RGB
    int16 min = MIN3( vRGB[0], vRGB[1], vRGB[2] );
    int i;
    for ( i = 0; i < 3; i++ ) {
        sRGB[i] = vRGB[i] - min;
    }
    
    return sRGB;
}

/**
 * Get color name from the Hue wheel in the HSV system
 * @see https://wisotop.de/hsv-und-hsl-farbmodell.php
 * @param h hue
 * @param s saturation
 * @return pointer to color name
 * @see colName
 */
char* getColorName( float h, float s ) {
    if ( s < SATURATION_THRESHOLD ) {        // less saturation => gray
        return colName[6];  // gray
    }
    
    // turn hue angle by 30 degree
    uint degr = ( (uint) ( h + 30. ) ) % 360;
    // get color sector from the hue wheel
    uint8 idx = degr / 60;
    return colName[idx];
}

/* [] END OF FILE */
